﻿// Copyright (c) .NET Foundation. All rights reserved.
// Licensed under the MIT License. See License.txt in the project root for license information.

using System;
using System.Collections.Concurrent;
using System.Threading;
using Microsoft.Azure.WebJobs.Host;
using Microsoft.Azure.WebJobs.Host.Executors;
using Microsoft.Azure.WebJobs.Host.Indexers;
using Microsoft.Azure.WebJobs.Host.Loggers;

namespace Microsoft.Azure.WebJobs
{
    /// <summary>
    /// Represents the configuration settings for a <see cref="JobHost"/>.
    /// </summary>
    public sealed class JobHostConfiguration : IServiceProvider
    {
        private static readonly IConsoleProvider ConsoleProvider = new DefaultConsoleProvider();

        private readonly DefaultStorageAccountProvider _storageAccountProvider;
        private readonly JobHostQueuesConfiguration _queueConfiguration = new JobHostQueuesConfiguration();
        private readonly JobHostTraceConfiguration _traceConfiguration = new JobHostTraceConfiguration();
        private readonly ConcurrentDictionary<Type, object> _services = new ConcurrentDictionary<Type, object>();
        private IJobHostContextFactory _contextFactory;

        // TEMP: This will go away in a future release (once ServiceBusConnectionString is removed from this class)
        private bool _serviceBusConnectionStringSet;
        private string _serviceBusConnectionString;

        private string _hostId;

        /// <summary>
        /// Initializes a new instance of the <see cref="JobHostConfiguration"/> class.
        /// </summary>
        public JobHostConfiguration()
            : this(null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="JobHostConfiguration"/> class, using the
        /// specified connection string for both reading and writing data as well as Dashboard logging.
        /// </summary>
        /// <param name="dashboardAndStorageConnectionString">The Azure Storage connection string to use.
        /// </param>
        public JobHostConfiguration(string dashboardAndStorageConnectionString)
        {
            if (!string.IsNullOrEmpty(dashboardAndStorageConnectionString))
            {
                _storageAccountProvider = new DefaultStorageAccountProvider(this, dashboardAndStorageConnectionString);
            }
            else
            {
                _storageAccountProvider = new DefaultStorageAccountProvider(this);
            }

            Singleton = new SingletonConfiguration();

            // add our built in services here
            IExtensionRegistry extensions = new DefaultExtensionRegistry();
            ITypeLocator typeLocator = new DefaultTypeLocator(ConsoleProvider.Out, extensions);
            AddService<IExtensionRegistry>(extensions);
            AddService<StorageClientFactory>(new StorageClientFactory());
            AddService<INameResolver>(new DefaultNameResolver());
            AddService<IJobActivator>(DefaultJobActivator.Instance);
            AddService<ITypeLocator>(typeLocator);
        }

        /// <summary>Gets or sets the host ID.</summary>
        /// <remarks>
        /// <para>
        /// All host instances that share the same host ID must be homogeneous. For example, they must use the same
        /// storage accounts and have the same list of functions. Host instances with the same host ID will scale out
        /// and share handling of work such as BlobTrigger and run from dashboard processing and providing a heartbeat
        /// to the dashboard indicating that an instance of the host running.
        /// </para>
        /// <para>
        /// If this value is <see langword="null"/>, a host ID will automatically be generated based on the assembly
        /// name of the first function.
        /// </para>
        /// <para>
        /// If non-homogeneous host instances share the same first function assembly,
        /// this property must be set explicitly; otherwise, the host instances will incorrectly try to share work as if
        /// they were homogeneous.
        /// </para>
        /// </remarks>
        public string HostId
        {
            get
            {
                return _hostId;
            }
            set
            {
                if (value != null && !HostIdValidator.IsValid(value))
                {
                    throw new ArgumentException(HostIdValidator.ValidationMessage, "value");
                }

                _hostId = value;
            }
        }

        /// <summary>Gets or sets the job activator.</summary>
        /// <remarks>The job activator creates instances of job classes when calling instance methods.</remarks>
        public IJobActivator JobActivator
        {
            get
            {
                return GetService<IJobActivator>();
            }
            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }
                AddService<IJobActivator>(value);
            }
        }

        /// <summary>
        /// Gets or sets the Azure Storage connection string used for logging and diagnostics.
        /// </summary>
        public string DashboardConnectionString
        {
            get { return _storageAccountProvider.DashboardConnectionString; }
            set { _storageAccountProvider.DashboardConnectionString = value; }
        }

        /// <summary>
        /// Gets or sets the Azure Storage connection string used for reading and writing data.
        /// </summary>
        public string StorageConnectionString
        {
            get { return _storageAccountProvider.StorageConnectionString; }
            set { _storageAccountProvider.StorageConnectionString = value; }
        }

        /// <summary>
        /// Gets or sets the Azure ServiceBus connection string.
        /// </summary>
        [Obsolete("Use ServiceBusConfiguration, and pass in via JobHostConfiguration.UseServiceBus.")]
        public string ServiceBusConnectionString
        {
            get
            {
                if (!_serviceBusConnectionStringSet)
                {
                    _serviceBusConnectionString = AmbientConnectionStringProvider.Instance.GetConnectionString(ConnectionStringNames.ServiceBus);
                    _serviceBusConnectionStringSet = true;
                }

                return _serviceBusConnectionString;
            }
            set
            {
                _serviceBusConnectionString = value;
                _serviceBusConnectionStringSet = true;
            }
        }

        /// <summary>
        /// Gets or sets the timeout that function invocations will be constrained to.
        /// <see cref="TimeoutAttribute"/> for details. If <see cref="TimeoutAttribute"/>
        /// is applied to a function or its containing class, that timeout value will override
        /// this global value.
        /// </summary>
        public TimeSpan? FunctionTimeout { get; set; }

        /// <summary>Gets or sets the type locator.</summary>
        public ITypeLocator TypeLocator
        {
            get
            {
                return GetService<ITypeLocator>();
            }
            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }
                AddService<ITypeLocator>(value);
            }
        }

        /// <summary>
        /// Gets or sets the name resolver used during indexing. 
        /// </summary>
        public INameResolver NameResolver
        {
            get
            {
                return GetService<INameResolver>();
            }
            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }
                AddService<INameResolver>(value);
            }
        }

        /// <summary>
        /// Gets the configuration used by <see cref="QueueTriggerAttribute"/>.
        /// </summary>
        public JobHostQueuesConfiguration Queues
        {
            get { return _queueConfiguration; }
        }

        /// <summary>
        /// Gets the configuration used by <see cref="SingletonAttribute"/>.
        /// </summary>
        public SingletonConfiguration Singleton
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the configuration for event tracing.
        /// </summary>
        public JobHostTraceConfiguration Tracing
        {
            get
            {
                return _traceConfiguration;
            }
        }

        private IJobHostContextFactory ContextFactory
        {
            get
            {
                if (_contextFactory == null)
                {
                    _contextFactory = new JobHostContextFactory(_storageAccountProvider, ConsoleProvider, this);
                }

                return _contextFactory;
            }
        }

        /// <summary>
        /// Gets or sets the <see cref="Host.StorageClientFactory"/> that will be used to create
        /// Azure Storage clients.
        /// </summary>
        [CLSCompliant(false)]
        public StorageClientFactory StorageClientFactory
        {
            get
            {
                return GetService<StorageClientFactory>();
            }
            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }
                AddService<StorageClientFactory>(value);
            }
        }

        /// <summary>Gets the service object of the specified type.</summary>
        /// <param name="serviceType">The type of service to get.</param>
        /// <returns>
        /// A service of the specified type, if one is available; otherwise, <see langword="null"/>.
        /// </returns>
        public object GetService(Type serviceType)
        {
            if (serviceType == null)
            {
                throw new ArgumentNullException("serviceType");
            }

            object service = null;
            _services.TryGetValue(serviceType, out service);

            if (service == null && serviceType == typeof(IJobHostContextFactory))
            {
                // ContextFactory must be delay created at the right time
                AddService<IJobHostContextFactory>(ContextFactory);
                return ContextFactory;
            }

            return service;
        }

        /// <summary>
        /// Gets the service object of the specified type.
        /// </summary>
        /// <typeparam name="TService">The type of service object to get.</typeparam>
        /// <returns>A service of the specified type, if one is available; otherwise, <see langword="null"/>.</returns>
        public TService GetService<TService>()
        {
            return (TService)this.GetService(typeof(TService));
        }

        /// <summary>
        /// Adds the specified service instance, replacing any existing service.
        /// </summary>
        /// <param name="serviceType">The service type</param>
        /// <param name="serviceInstance">The service instance</param>
        public void AddService(Type serviceType, object serviceInstance)
        {
            if (serviceType == null)
            {
                throw new ArgumentNullException("serviceType");
            }
            if (!serviceType.IsAssignableFrom(serviceInstance.GetType()))
            {
                throw new ArgumentOutOfRangeException("serviceInstance");
            }

            _services.AddOrUpdate(serviceType, serviceInstance, (key, existingValue) =>
                {
                    // always replace existing values
                    return serviceInstance;
                });
        }

        /// <summary>
        /// Adds the specified service instance, replacing any existing service.
        /// </summary>
        /// <typeparam name="TService">The service type</typeparam>
        /// <param name="serviceInstance">The service instance</param>
        public void AddService<TService>(TService serviceInstance)
        {
            AddService(typeof(TService), serviceInstance);
        }
    }
}
